/*
 * Copyright (C) 2007 Markus Niemann
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * contact information see:
 * http://www.mpipks-dresden.mpg.de/cgi-bin/user-info?user=niemann
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <glib.h>

#define BILLIARDRADIUS 0.5
#define MIN_BILLIARDRADIUS 0.05
#define MAX_BILLIARDRADIUS 0.99
#define BILLIARDSIZE 400
#define INFORMATIONTABLESIZE 50
#define EPSILON 1.0e-7
#define MS_BETWEEN_FRAMES 50
#define COLORDEPTH 24
#define MIN_MOVEMENT (1.0/(double)(BILLIARDSIZE))
#define SPEED_FACTOR 0.5
#define NUMBER_OF_COLORS 50
#define ITERATIONS_BEFORE_HISTOGRAM_UPDATE 10

/* conditioned compilation */
#define DISPLAY_RADIUS_SLIDER
#undef BILLIARD_REFLECTING_BORDERS
#define HISTOGRAM_USE_EXPECTATION_VALUE 
/* otherwise it uses the maximal value to normalise the histogram */

/* define one of the following to set the color coding of the histogram */
#define HISTOGRAM_BLUE_GREEN_RED
#undef HISTOGRAM_WHITE_BLACK
#undef HISTOGRAM_GREEN_RED

/* text used for buttons, etc. */
#define TEXT_TITLE_WINDOW "Sinai Billard"
#define TEXT_BUTTON_START "Start"
#define TEXT_LABEL_SLIDER_SPEED "Geschwindigkeit:"
#define TEXT_BUTTON_MAXIMUM_SPEED "hohe Geschwindigkeit"
#define TEXT_BUTTON_HISTOGRAM "Histogramm"
#define TEXT_BUTTON_PRESET "Preset"
#define TEXT_BUTTON_QUIT "Ende"
#define TEXT_LABEL_SLIDER_RADIUS "Radius:"


/* structure containing the information for one Sinai billiard */
struct sinaiBilliard
{
    double xpos, ypos;
    double xdir, ydir;
    double radius;
    unsigned int sidelength;  /* pixel */
    unsigned int besidelength;
    double histogram[BILLIARDSIZE+1][BILLIARDSIZE+1];
    double max_histogram;  /* the maximal value in the histogram */
    double total_length; /* the total length of the trajectory in the histogram */
    GtkWidget *canvas;
    GdkPixmap *buffer;
    int histogram_update_counter;
};

   
/* global variables */
GtkWidget *mainWindow;
struct sinaiBilliard leftBilliard, rightBilliard;
gboolean animation_running = FALSE;
gboolean animation_maximum_speed = FALSE;
gboolean histogram_active = FALSE;
gint tag_function_animation;
gint tag_function_idle;
double animation_movement = MIN_MOVEMENT;
GdkGC *histogram_gc = NULL;
GdkColor my_colors[NUMBER_OF_COLORS];


/* draw a (extended) point on the position of the ball */
void billiard_draw_position (
    struct sinaiBilliard *billiard )
{
    int x = (int)(0.5 * (1.0 + billiard->xpos)
              * (double)(billiard->sidelength));
    int y = (int)(0.5 * (1.0 + billiard->ypos)
              * (double)(billiard->sidelength));

    gdk_draw_point (billiard->buffer,
         billiard->canvas->style->black_gc, x, y);

    if (x > 0) gdk_draw_point (billiard->buffer,
                   billiard->canvas->style->black_gc, x - 1, y);
    if (y > 0) gdk_draw_point (billiard->buffer,
                   billiard->canvas->style->black_gc, x, y - 1);
    if (x < BILLIARDSIZE - 1) gdk_draw_point (billiard->buffer,
                   billiard->canvas->style->black_gc, x + 1, y);
    if (y < BILLIARDSIZE - 1) gdk_draw_point (billiard->buffer,
                   billiard->canvas->style->black_gc, x, y + 1);
}

/* draws the basic components (i.e. frame, scatterer and ball) of the
   billiard */
void billiard_draw_basic_sinai (
    struct sinaiBilliard *billiard )
{
    gdk_draw_rectangle (billiard->buffer,
        billiard->canvas->style->white_gc, TRUE,
        0, 0, billiard->sidelength, billiard->sidelength);
    gdk_draw_rectangle (billiard->buffer,
        billiard->canvas->style->black_gc, FALSE,
        0, 0, billiard->sidelength-1, billiard->sidelength-1);
    { /* draw the scatterer */
        int corner = (int)((1.0 - billiard->radius) 
                     * 0.5 * (double)(billiard->sidelength));
        int width = (int)(billiard->radius * (double)(billiard->sidelength));

        gdk_draw_arc (billiard->buffer,
            billiard->canvas->style->black_gc, TRUE,
            corner, corner, width, width, 0, 64*360 );
    }

    if (!animation_running) billiard_draw_position (billiard);
}



/* draws the billiard with the histogramm as a ground */
void billiard_draw_sinai_with_histogram (
    struct sinaiBilliard *billiard,
    gboolean draw_now )
{
    int i, j;
#ifndef HISTOGRAM_USE_EXPECTATION_VALUE
    double prefactor = (double)(NUMBER_OF_COLORS) / billiard->max_histogram;
#else
    double prefactor = (double)(NUMBER_OF_COLORS) * (double)(BILLIARDSIZE) *
                       (double)(BILLIARDSIZE) * 0.5 / billiard->total_length;
#endif

    /* update slower, because it will get stuck otherwise */
    billiard->histogram_update_counter++;
    if ((billiard->histogram_update_counter<ITERATIONS_BEFORE_HISTOGRAM_UPDATE)
        && !draw_now) return;
    billiard->histogram_update_counter = 0;

    for (i=0; i<BILLIARDSIZE; i++)
    for (j=0; j<BILLIARDSIZE; j++)
    {
        int value;

        value = (int)(prefactor * billiard->histogram[i][j]);
        if (value >= NUMBER_OF_COLORS) value = NUMBER_OF_COLORS - 1;

        gdk_gc_set_foreground (histogram_gc, my_colors+value);
        gdk_draw_point (billiard->buffer, histogram_gc, i, j);
    }

    gdk_draw_rectangle (billiard->buffer,
        billiard->canvas->style->black_gc, FALSE,
        0, 0, billiard->sidelength-1, billiard->sidelength-1);
    { /* draw the scatterer */
        int corner = (int)((1.0 - billiard->radius)
                     * 0.5 * (double)(billiard->sidelength));
        int width = (int)(billiard->radius * (double)(billiard->sidelength));

        gdk_draw_arc (billiard->buffer,
            billiard->canvas->style->black_gc, TRUE,
            corner, corner, width, width, 0, 64*360 );
    }
}


/* sets up the color scheme for the histogram */
void histogram_set_up_colors ()
{
    int i;
    GdkColormap *colormap;
    gboolean success[NUMBER_OF_COLORS];
    gint result;

    /* dirty hack... */
    histogram_gc = gdk_gc_new (leftBilliard.buffer);

    colormap = gdk_colormap_get_system ();

    for (i=0; i<NUMBER_OF_COLORS; i++) my_colors[i].pixel = i;

#if defined HISTOGRAM_BLUE_GREEN_RED
    for (i=0; i<NUMBER_OF_COLORS/2; i++)
    {
        my_colors[i].green = (guint16)(65535.0 * 2.0 * (double)(i) 
                                     / (double)(NUMBER_OF_COLORS));
        my_colors[i].blue = (guint16)(65535.0 * (1.0 - 2.0 * (double)(i)
                                     / (double)(NUMBER_OF_COLORS)));
        my_colors[i].red = 0;
    }

    for (i=NUMBER_OF_COLORS/2; i<NUMBER_OF_COLORS; i++)
    {
        my_colors[i].red = (guint16)(65535.0 * 2.0 * 
           (double)(i - NUMBER_OF_COLORS/2) / (double)(NUMBER_OF_COLORS));
        my_colors[i].green = (guint16)(65535.0 * (1.0 - 2.0 * 
           (double)(i - NUMBER_OF_COLORS/2) / (double)(NUMBER_OF_COLORS)));
        my_colors[i].blue = 0;
    }
#elif defined HISTOGRAM_WHITE_BLACK
    for (i=0; i<NUMBER_OF_COLORS; i++)
    {
        my_colors[i].red = (guint16)(65535.0 * (double)(NUMBER_OF_COLORS - i) / 
                              (double)(NUMBER_OF_COLORS));
        my_colors[i].green = my_colors[i].red;
        my_colors[i].blue = my_colors[i].red;
    }
#elif defined HISTOGRAM_GREEN_RED
    for (i=0; i<NUMBER_OF_COLORS; i++)
    {
        my_colors[i].green = (guint16)(65535.0 * (double)(NUMBER_OF_COLORS - i)
                               / (double)(NUMBER_OF_COLORS));
        my_colors[i].red = (guint16)(65535.0 * (double)(i) /
                              (double)(NUMBER_OF_COLORS));
        my_colors[i].blue = 0;
    }
#else
#error no color scheme defined for histogram
#endif

    result = gdk_colormap_alloc_colors (colormap, my_colors,
             NUMBER_OF_COLORS, TRUE, TRUE, success);
    if (result != 0) printf ("%i colors could not be allocated\n", result);

    gdk_gc_set_colormap (histogram_gc, colormap);
}


/* draws the rose indication the direction */
void billiard_draw_rose (
    struct sinaiBilliard *billiard )
{
    int x_middle = billiard->sidelength + billiard->besidelength / 2;
    int y_middle = billiard->besidelength / 2;
    int half_width = billiard->besidelength / 2 - 5;

    gdk_draw_rectangle (billiard->buffer,
        billiard->canvas->style->white_gc, TRUE,
        x_middle - half_width, y_middle - half_width, 
        2 * half_width, 2 * half_width); 
    gdk_draw_rectangle (billiard->buffer,
        billiard->canvas->style->black_gc, FALSE,
        x_middle - half_width, y_middle - half_width, 
        2 * half_width, 2 * half_width); 
    { /* draw the line indicating the direction */
        int x_end = x_middle + (int)(billiard->xdir * (double)(half_width));
        int y_end = y_middle + (int)(billiard->ydir * (double)(half_width));

        gdk_draw_line (billiard->buffer,
            billiard->canvas->style->black_gc,
            x_middle, y_middle, x_end, y_end);
    }
}


/* draws the colors for the histogram on the right side of the
   screen */
void histogram_draw_information_panel (
    struct sinaiBilliard *billiard)
{
    int i;
    int left = BILLIARDSIZE + 5;
    int right = BILLIARDSIZE + INFORMATIONTABLESIZE - 5;
    int down = BILLIARDSIZE - 5;
    int up = BILLIARDSIZE - (6 + NUMBER_OF_COLORS);

    for (i=0; i < NUMBER_OF_COLORS; i++)
    {
        gdk_gc_set_foreground (histogram_gc, my_colors + i);
        gdk_draw_line (billiard->buffer, histogram_gc,
            left, down - i - 1, right, down - i - 1);
    }
    gdk_draw_rectangle (billiard->buffer,
        billiard->canvas->style->black_gc,
        FALSE, left, up, right - left, down - up);
}


/* erases the information panel of the histogram */
void histogram_erase_information_panel (
    struct sinaiBilliard *billiard)
{
    int left = BILLIARDSIZE + 5;
    int right = BILLIARDSIZE + INFORMATIONTABLESIZE - 5;
    int down = BILLIARDSIZE - 5;
    int up = BILLIARDSIZE - (6 + NUMBER_OF_COLORS);

    gdk_draw_rectangle (billiard->buffer,
        billiard->canvas->style->bg_gc[GTK_WIDGET_STATE(billiard->canvas)],
        TRUE, left, up, right - left + 1, down - up + 1);
}



/* sets up the screen for a billiard, including the parts
   which stay the same all the time */
void billiard_set_up_screen (
    struct sinaiBilliard *billiard)
{
    gdk_draw_rectangle (billiard->buffer,
        billiard->canvas->style->bg_gc[GTK_WIDGET_STATE(billiard->canvas)],
        TRUE, 0, 0, billiard->sidelength+billiard->besidelength, 
        billiard->sidelength);
    billiard_draw_basic_sinai (billiard);
    billiard_draw_rose (billiard);
}


/* initiates a redrawing of the screen */
void initiate_redraw ()
{
    gtk_widget_queue_draw_area (GTK_WIDGET (leftBilliard.canvas), 0, 0,
          leftBilliard.sidelength + leftBilliard.besidelength,
          leftBilliard.sidelength);
    gtk_widget_queue_draw_area (GTK_WIDGET (rightBilliard.canvas), 0, 0,
          rightBilliard.sidelength + rightBilliard.besidelength,
          rightBilliard.sidelength);
}


/* normalises the direction vector of a billiard */
void normalise_direction (struct sinaiBilliard *billiard)
{
    double length;

    length = hypot (billiard->xdir, billiard->ydir);
    if (length >= EPSILON)
    {
        billiard->xdir /= length;
        billiard->ydir /= length;
    }
}


/* set preset values for the two billiards */
/* requires radius to be set */
void set_billiard_preset ()
{
    leftBilliard.xpos = 1.0 - 0.5 * (1.0 - leftBilliard.radius);
    leftBilliard.ypos = 0.0;
    leftBilliard.xdir = 1.0;
    leftBilliard.ydir = 0.1;
    normalise_direction (&leftBilliard);

    rightBilliard.xpos = 1.0 - 0.5 * (1.0 - rightBilliard.radius);
    rightBilliard.ypos = 0.05;
    rightBilliard.xdir = 1.0;
    rightBilliard.ydir = 0.1;
    normalise_direction (&rightBilliard);
}


/* clears the histograms */
void clear_billiard_histogram ()
{
    int i, j;

    for (i=0; i<=BILLIARDSIZE; i++)
    for (j=0; j<=BILLIARDSIZE; j++)
        leftBilliard.histogram[i][j] = 0.0;
    for (i=0; i<=BILLIARDSIZE; i++)
    for (j=0; j<=BILLIARDSIZE; j++)
        rightBilliard.histogram[i][j] = 0.0;

    i = (int)(0.5*(leftBilliard.xpos + 1.0)*(double)(leftBilliard.sidelength));
    j = (int)(0.5*(leftBilliard.ypos + 1.0)*(double)(leftBilliard.sidelength));
    leftBilliard.histogram[i][j] = EPSILON;
    leftBilliard.max_histogram = EPSILON;
    leftBilliard.total_length = EPSILON;

    i =(int)(0.5*(rightBilliard.xpos + 1.0)*(double)(rightBilliard.sidelength));
    j =(int)(0.5*(rightBilliard.ypos + 1.0)*(double)(rightBilliard.sidelength));
    rightBilliard.histogram[i][j] = EPSILON;
    rightBilliard.max_histogram = EPSILON;
    rightBilliard.total_length = EPSILON;
}


/* handles a mouse click in the billiards */
gboolean billiard_button_pressed_event (
    GtkWidget *widget,
    GdkEventButton *event,
    gpointer side )
{
    struct sinaiBilliard *billiard;
    int x = (unsigned int)(event->x);
    int y = (unsigned int)(event->y);

    if (strcmp ((char *)(side), "left") == 0) billiard = &leftBilliard;
    else billiard = &rightBilliard; 

    if (x < BILLIARDSIZE)
    {   /* inside the billiard area */
        double x_coor = 2.0 * ( (double)(x) / (double)(BILLIARDSIZE) - 0.5);
        double y_coor = 2.0 * ( (double)(y) / (double)(BILLIARDSIZE) - 0.5);

        /* check if it outside the scatterer */
        if ( x_coor * x_coor + y_coor * y_coor > 
             billiard->radius * billiard->radius )
        {
            billiard->xpos = x_coor;
            billiard->ypos = y_coor;
            billiard_draw_basic_sinai (billiard);
            initiate_redraw();
        }
    } else {  /* outside billiard area */
        x -= BILLIARDSIZE;

        /* check for being inside the direction area */
        if (y < INFORMATIONTABLESIZE)
        { /* set new direction */
            x -= INFORMATIONTABLESIZE / 2;
            y -= INFORMATIONTABLESIZE / 2;

            if ((x != 0) && (y != 0))
            {
                billiard->xdir = (double)(x);
                billiard->ydir = (double)(y);
                normalise_direction (billiard);
                billiard_draw_rose (billiard);
                initiate_redraw();
            }
        }
    }

    return TRUE;
}


/* handles the exposure event (i.e. redraw...) */
gboolean billiard_exposure_event (
    GtkWidget *widget, 
    GdkEventExpose *event,
    gpointer info )
{
    struct sinaiBilliard *billiard;

    if (strcmp((char *)info, "left") == 0) billiard = &leftBilliard;
    else billiard = &rightBilliard;

    /* from GTK tutorial */
    gdk_draw_drawable(widget->window,
                      widget->style->fg_gc[GTK_WIDGET_STATE (widget)],
                      billiard->buffer,
                      event->area.x, event->area.y,
                      event->area.x, event->area.y,
                      event->area.width, event->area.height);

    return FALSE;
}


/* sets up the two billiards */
void set_up_billiards ()
{
    leftBilliard.radius = BILLIARDRADIUS;
    rightBilliard.radius = BILLIARDRADIUS;
    leftBilliard.sidelength = BILLIARDSIZE;
    rightBilliard.sidelength = BILLIARDSIZE;
    leftBilliard.besidelength = INFORMATIONTABLESIZE;
    rightBilliard.besidelength = INFORMATIONTABLESIZE;
    leftBilliard.histogram_update_counter = 0;
    rightBilliard.histogram_update_counter = 0;
    leftBilliard.total_length = EPSILON;
    rightBilliard.total_length = EPSILON;
    set_billiard_preset ();
    clear_billiard_histogram ();

    leftBilliard.canvas = gtk_drawing_area_new ();
    gtk_widget_set_size_request (GTK_WIDGET (leftBilliard.canvas),
        leftBilliard.sidelength + leftBilliard.besidelength,
        leftBilliard.sidelength);
    leftBilliard.buffer = gdk_pixmap_new (
        GDK_DRAWABLE(leftBilliard.canvas->window),
        leftBilliard.sidelength + leftBilliard.besidelength,
        leftBilliard.sidelength, COLORDEPTH);
    gtk_signal_connect (GTK_OBJECT (leftBilliard.canvas), "expose_event",
        (GtkSignalFunc) billiard_exposure_event, "left");
    gtk_signal_connect (GTK_OBJECT (leftBilliard.canvas), "button_press_event",
        (GtkSignalFunc) billiard_button_pressed_event, "left");
    gtk_widget_set_events (leftBilliard.canvas, GDK_EXPOSURE_MASK
                               | GDK_BUTTON_PRESS_MASK );

    rightBilliard.canvas = gtk_drawing_area_new ();
    gtk_drawing_area_size (GTK_DRAWING_AREA (rightBilliard.canvas),
        rightBilliard.sidelength + rightBilliard.besidelength,
        rightBilliard.sidelength);
    rightBilliard.buffer = gdk_pixmap_new (
        GDK_DRAWABLE(rightBilliard.canvas->window),
        rightBilliard.sidelength + rightBilliard.besidelength,
        rightBilliard.sidelength, COLORDEPTH);
    gtk_signal_connect (GTK_OBJECT (rightBilliard.canvas), "expose_event",
        (GtkSignalFunc) billiard_exposure_event, "right");
    gtk_signal_connect (GTK_OBJECT (rightBilliard.canvas), "button_press_event",
        (GtkSignalFunc) billiard_button_pressed_event, "right");
    gtk_widget_set_events (rightBilliard.canvas, GDK_EXPOSURE_MASK
                               | GDK_BUTTON_PRESS_MASK );
}


/* update the histogram by considering the movement of the momentarily
   position to a given endpoint */
void billiard_add_to_histogram (
    struct sinaiBilliard *billiard,
    double length )
{
    int x_step = +1;
    int y_step = +1;
    int x_bin, y_bin;
    double length_to_do = length;
    double x_large = billiard->xpos * 0.5 * (double)(BILLIARDSIZE);
    double y_large = billiard->ypos * 0.5 * (double)(BILLIARDSIZE);
    double x_ldir = billiard->xdir * 0.5 * (double)(BILLIARDSIZE);
    double y_ldir = billiard->ydir * 0.5 * (double)(BILLIARDSIZE);
    double length_to_top;
    double length_to_right;

    billiard->total_length += length;

    if (x_ldir >= 0.0)
    {
        x_large += 0.5 * (double)(BILLIARDSIZE);
        x_bin = (int)(floor(x_large));
        x_large -= floor(x_large);
    } else {
        x_step = -1;
        x_ldir = -x_ldir;
        x_large = -x_large;
        x_large += 0.5 * (double)(BILLIARDSIZE);
        x_bin = BILLIARDSIZE - 1 - (int)(floor(x_large));
        x_large -= floor(x_large);
    }
    if (y_ldir >= 0.0)
    {
        y_large += 0.5 * (double)(BILLIARDSIZE);
        y_bin = (int)(floor(y_large));
        y_large -= floor(y_large);
    } else {
        y_step = -1;
        y_ldir = -y_ldir;
        y_large = -y_large;
        y_large += 0.5 * (double)(BILLIARDSIZE);
        y_bin = BILLIARDSIZE - 1 - (int)(floor(y_large));
        y_large -= floor(y_large);
    }

    /* start of loop */
    length_to_top = (1.0 - x_large) / x_ldir;
    length_to_right = (1.0 - y_large) / y_ldir;

    if (!isnormal(length_to_top)) length_to_top = 1.e7; 
    if (!isnormal(length_to_right)) length_to_right = 1.e7; 

    while ((length_to_do > length_to_top) && (length_to_do > length_to_right))
    {
        if (length_to_top < length_to_right)
        {  /* getting out of the top */
            x_large = 0.0;
            y_large += length_to_top * y_ldir;

            billiard->histogram[x_bin][y_bin] += length_to_top;
            if (billiard->histogram[x_bin][y_bin] > billiard->max_histogram)
                billiard->max_histogram = billiard->histogram[x_bin][y_bin];

            x_bin += x_step;
            length_to_do -= length_to_top;
        } else {
            y_large = 0.0;
            x_large += length_to_right * x_ldir;

            billiard->histogram[x_bin][y_bin] += length_to_right;
            if (billiard->histogram[x_bin][y_bin] > billiard->max_histogram)
                billiard->max_histogram = billiard->histogram[x_bin][y_bin];

            y_bin += y_step;
            length_to_do -= length_to_right;
        }

        length_to_top = (1.0 - x_large) / x_ldir;
        length_to_right = (1.0 - y_large) / y_ldir;

        if (!isnormal(length_to_top)) length_to_top = 1.e7;
        if (!isnormal(length_to_right)) length_to_right = 1.e7;
    } /* while */

    billiard->histogram[x_bin][y_bin] += length_to_do;
    if (billiard->histogram[x_bin][y_bin] > billiard->max_histogram)
        billiard->max_histogram = billiard->histogram[x_bin][y_bin];

}


/* draws a line from the momentarily position to the endposition
   in the billiard */
void billiard_draw_line (
    struct sinaiBilliard *billiard,
    double xend,
    double yend )
{
    int x1 = (int)(0.5 * (1.0 + billiard->xpos)
             * (double)(billiard->sidelength));
    int y1 = (int)(0.5 * (1.0 + billiard->ypos)
             * (double)(billiard->sidelength));
    int x2 = (int)(0.5 * (1.0 + xend)
             * (double)(billiard->sidelength));
    int y2 = (int)(0.5 * (1.0 + yend)
             * (double)(billiard->sidelength));

    if (x1 == BILLIARDSIZE) x1--;
    if (y1 == BILLIARDSIZE) y1--;
    if (x2 == BILLIARDSIZE) x2--;
    if (y2 == BILLIARDSIZE) y2--;

    gdk_draw_line (billiard->buffer,
        billiard->canvas->style->black_gc, x1, y1, x2, y2);
}


/* do one step for one billiard and draw lines resp. updates
   the histogram */
void billiard_animation_step (
    struct sinaiBilliard *billiard,
    gboolean paint ) /* should I paint the lines */
{
    double endpoint_x, endpoint_y;
    double new_xpos, new_ypos;
    double new_xdir, new_ydir;
    double length_to_do = animation_movement;
    double next_obstacle;
    double length_compare;

    while (length_to_do > EPSILON * animation_movement)
    {
        /* set up for free flight */
        next_obstacle = length_to_do;
        endpoint_x = new_xpos = billiard->xpos + length_to_do * billiard->xdir;
        endpoint_y = new_ypos = billiard->ypos + length_to_do * billiard->ydir;
        new_xdir = billiard->xdir;
        new_ydir = billiard->ydir;

        /* test for left boundary */
        if (billiard->xdir < 0.0)
        {
            length_compare = - (billiard->xpos + 1.0) / billiard->xdir;
            if (length_compare < next_obstacle)
            {
                next_obstacle = length_compare;
                endpoint_x = -1.0;
                endpoint_y = billiard->ypos + next_obstacle * billiard->ydir;
                new_ypos = endpoint_y;
#ifndef BILLIARD_REFLECTING_BORDERS
                new_xpos = 1.0;
#else
                new_xpos = -1.0;
                new_xdir = -billiard->xdir;
#endif
            }
        }

        /* test for right boudary */
        if (billiard->xdir > 0.0)
        {
            length_compare = (1.0 - billiard->xpos) / billiard->xdir;
            if (length_compare < next_obstacle)
            {
                next_obstacle = length_compare;
                endpoint_x = 1.0;
                endpoint_y = billiard->ypos + next_obstacle * billiard->ydir;
                new_ypos = endpoint_y;
#ifndef BILLIARD_REFLECTING_BORDERS
                new_xpos = -1.0;
#else
                new_xpos = 1.0;
                new_xdir = -billiard->xdir;
#endif
            }
        }

        /* test for upper boundary */
        if (billiard->ydir < 0.0)
        {
            length_compare = - (billiard->ypos + 1.0) / billiard->ydir;
            if (length_compare < next_obstacle)
            {
                next_obstacle = length_compare;
                endpoint_x = billiard->xpos + next_obstacle * billiard->xdir;
                new_xpos = endpoint_x;
                endpoint_y = -1.0;
#ifndef BILLIARD_REFLECTING_BORDERS
                new_ypos = 1.0;
#else
                new_ypos = -1.0;
                new_ydir = -billiard->ydir;
#endif
            }
        }

        /* test for lower boundary */
        if (billiard->ydir > 0.0)
        {
            length_compare = (1.0 - billiard->ypos) / billiard->ydir;
            if (length_compare < next_obstacle)
            {
                next_obstacle = length_compare;
                endpoint_x = billiard->xpos + next_obstacle * billiard->xdir;
                new_xpos = endpoint_x;
                endpoint_y = 1.0;
#ifndef BILLIARD_REFLECTING_BORDERS
                new_ypos = -1.0;
#else
                new_ypos = 1.0;
                new_ydir = -billiard->ydir;
#endif
            }
        }

        /* test for scatterer */
        length_compare = - (billiard->xpos * billiard->xdir +
                billiard->ypos * billiard->ydir);
        if (length_compare > 0.0) /* the nearest encounter of the origin */
        {                         /* is in the future */
            double distance_sqr = billiard->xpos * billiard->xpos +
                                  billiard->ypos * billiard->ypos;
            double encounter_sqr = distance_sqr -
                                   length_compare * length_compare;
            if (encounter_sqr < billiard->radius * billiard->radius)
            {     /* we actually hit and have to subtract the length */
                  /* inside the scatterer */
                length_compare -= sqrt (billiard->radius * billiard->radius -
                                        encounter_sqr);
                if (length_compare < next_obstacle)  /* now it is sure */
                {
                    double prefactor;

                    next_obstacle = length_compare;
                    endpoint_x = billiard->xpos + next_obstacle * 
                                                  billiard->xdir;
                    new_xpos = endpoint_x;
                    endpoint_y = billiard->ypos + next_obstacle * 
                                                  billiard->ydir;
                    new_ypos = endpoint_y;
                    /* mirror the direction */
                    prefactor = 2.0 * (billiard->xdir * endpoint_x +
                                      billiard->ydir * endpoint_y) / 
                                   (billiard->radius * billiard->radius);
                    new_xdir = billiard->xdir - prefactor * endpoint_x;
                    new_ydir = billiard->ydir - prefactor * endpoint_y;
                }
            }
        }
        if (histogram_active) 
            billiard_add_to_histogram (billiard, next_obstacle);
        else if (paint)
            billiard_draw_line (billiard, endpoint_x, endpoint_y);

        billiard->xpos = new_xpos;
        billiard->ypos = new_ypos;
        billiard->xdir = new_xdir;
        billiard->ydir = new_ydir;
        normalise_direction (billiard);

        length_to_do -= next_obstacle;
    }
}


/* performing one animation step with drawing */
gint animation_step (
    gpointer data )
{
    if (!histogram_active) 
    {
        billiard_draw_basic_sinai (&leftBilliard);
        billiard_draw_basic_sinai (&rightBilliard);
    }
    billiard_animation_step (&leftBilliard, TRUE);
    billiard_animation_step (&rightBilliard, TRUE);
    if (histogram_active)
    {
        billiard_draw_sinai_with_histogram (&leftBilliard, FALSE);
        billiard_draw_sinai_with_histogram (&rightBilliard, FALSE);
    } else {
        billiard_draw_position (&leftBilliard);
        billiard_draw_position (&rightBilliard);
    }
    billiard_draw_rose (&leftBilliard);
    billiard_draw_rose (&rightBilliard);
    initiate_redraw ();
    return TRUE;
}


/* doing the animation in maximum speed without drawing in the idle times */
gint animation_step_in_idle (
    gpointer data )
{
    billiard_animation_step (&leftBilliard, FALSE);
    billiard_animation_step (&rightBilliard, FALSE);

    return TRUE;
}


/* starts the animation */
void animation_start ()
{
    if (animation_running) return;
   
    animation_running = TRUE;
    tag_function_animation = g_timeout_add (MS_BETWEEN_FRAMES,
        (GtkFunction)animation_step, NULL);

    if (animation_maximum_speed)
    {
        tag_function_idle = gtk_idle_add ( 
            (GtkFunction)animation_step_in_idle, NULL);
    }
}


/* stops the animation */
void animation_stop ()
{
    if (!animation_running) return;

    animation_running = FALSE;
    g_source_remove (tag_function_animation);
    if (animation_maximum_speed)
    {
        gtk_idle_remove (tag_function_idle);
    }
}


/* handles the buttons on the bottom of the screen */
void button_pressed ( 
    GtkWidget *widget, 
    gpointer callback_data )
{
    char *button_name = (char *)(callback_data);

    if (strcmp (button_name, "start") == 0) 
    {
        if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget)))
        {
            animation_start ();
        } else {
            animation_stop ();

            /* make sure everythink is up to date */
            if (histogram_active)
            {
                billiard_draw_sinai_with_histogram (&leftBilliard, TRUE);
                billiard_draw_sinai_with_histogram (&rightBilliard, TRUE);
            } else {
                billiard_draw_basic_sinai (&leftBilliard);
                billiard_draw_basic_sinai (&rightBilliard);
            }
            billiard_draw_rose (&leftBilliard);
            billiard_draw_rose (&rightBilliard);
            initiate_redraw ();
        }
    }

    if (strcmp (button_name, "maximum_speed") == 0) 
    {
        gboolean animation_start_stop = animation_running;

        if (animation_start_stop) animation_stop();
        if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget)))
        {
            animation_maximum_speed = TRUE;
        } else {
            animation_maximum_speed = FALSE;
        }
        if (animation_start_stop) animation_start();
    }

    if (strcmp (button_name, "histogram") == 0) 
    {
        if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget)))
        {
            clear_billiard_histogram ();
            histogram_active = TRUE;
            billiard_draw_sinai_with_histogram (&leftBilliard, TRUE);
            billiard_draw_sinai_with_histogram (&rightBilliard, TRUE);
            histogram_draw_information_panel (&leftBilliard);
            histogram_draw_information_panel (&rightBilliard);
            initiate_redraw ();
        } else {
            histogram_active = FALSE;
            billiard_draw_basic_sinai (&leftBilliard);
            billiard_draw_basic_sinai (&rightBilliard);
            histogram_erase_information_panel (&leftBilliard);
            histogram_erase_information_panel (&rightBilliard);
            initiate_redraw ();
        }
    }

    if (strcmp (button_name, "preset") == 0) 
    {
        set_billiard_preset ();
        billiard_draw_basic_sinai (&leftBilliard);
        billiard_draw_basic_sinai (&rightBilliard);
        billiard_draw_rose (&leftBilliard);
        billiard_draw_rose (&rightBilliard);
        initiate_redraw ();
    }
}


void slider_speed_changed (
    GtkScale *scale,
    gpointer user_data)
{
    double value = (double)(gtk_range_get_value(GTK_RANGE(scale)));
    gboolean animation_start_stop = animation_running;

    if (animation_start_stop) animation_stop();
    animation_movement = MIN_MOVEMENT + SPEED_FACTOR * value;
    if (animation_start_stop) animation_start();
}


void slider_radius_changed (
    GtkScale *scale,
    gpointer user_data)
{
    double value = (double)(gtk_range_get_value(GTK_RANGE(scale)));

    leftBilliard.radius = value;
    rightBilliard.radius = value;

    set_billiard_preset ();

    if (histogram_active)
     {
         billiard_draw_sinai_with_histogram (&leftBilliard, TRUE);
         billiard_draw_sinai_with_histogram (&rightBilliard, TRUE);
     } else {
         billiard_draw_basic_sinai (&leftBilliard);
         billiard_draw_basic_sinai (&rightBilliard);
     }
     initiate_redraw ();
}


int main (int argc, char *argv[])
{
    GtkWidget *button;
    GtkWidget *verticalBox;
    GtkWidget *horizontalBox;
    GtkWidget *separator;
    GtkWidget *slider;
    GtkWidget *label;
    
    gtk_init (&argc, &argv);

    /* do the stuff for the main window */
    mainWindow = gtk_window_new (GTK_WINDOW_TOPLEVEL);
    gtk_window_set_title (GTK_WINDOW (mainWindow), TEXT_TITLE_WINDOW);
    g_signal_connect (G_OBJECT (mainWindow), "delete_event",
                      G_CALLBACK (gtk_main_quit), NULL);

    /* start filling the boxes, first the vertical one */
    verticalBox = gtk_vbox_new (FALSE, 0);

    /* do the first horizontal box with the billiards */
    horizontalBox = gtk_hbox_new (FALSE, 10);

    set_up_billiards ();
    gtk_box_pack_start (GTK_BOX (horizontalBox), leftBilliard.canvas,
                        FALSE, FALSE, 10);
    gtk_widget_show (leftBilliard.canvas);
    gtk_box_pack_start (GTK_BOX (horizontalBox), rightBilliard.canvas,
                        FALSE, FALSE, 10);
    gtk_widget_show (rightBilliard.canvas);

    gtk_box_pack_start (GTK_BOX (verticalBox), horizontalBox, FALSE, FALSE, 10);
    gtk_widget_show (horizontalBox);

    separator = gtk_hseparator_new ();
    gtk_box_pack_start (GTK_BOX (verticalBox), separator, FALSE, TRUE, 0);
    gtk_widget_show (separator);

    horizontalBox = gtk_hbox_new (FALSE, 10);

    /* insert buttons */
    label = gtk_label_new ("");
    gtk_box_pack_start (GTK_BOX (horizontalBox), label, FALSE, FALSE, 0);
    gtk_widget_show (label);

    button = gtk_toggle_button_new_with_label (TEXT_BUTTON_START);
    g_signal_connect (G_OBJECT (button), "clicked", 
                      G_CALLBACK (button_pressed), "start");
    gtk_box_pack_start (GTK_BOX (horizontalBox), button, FALSE, FALSE, 0);
    gtk_widget_show (button);

    label = gtk_label_new (TEXT_LABEL_SLIDER_SPEED);
    gtk_box_pack_start (GTK_BOX (horizontalBox), label, FALSE, FALSE, 0);
    gtk_widget_show (label);

    slider = gtk_hscale_new_with_range (0.0, 1.0, 0.1);
    g_signal_connect (G_OBJECT (slider), "value_changed",
                      G_CALLBACK (slider_speed_changed), NULL);
    gtk_scale_set_draw_value (GTK_SCALE (slider), FALSE);
    gtk_widget_set_size_request (GTK_WIDGET (slider), 100, -1);
    gtk_range_set_update_policy (GTK_RANGE(slider), GTK_UPDATE_DISCONTINUOUS);
    gtk_box_pack_start (GTK_BOX (horizontalBox), slider, FALSE, FALSE, 0);
    gtk_widget_show (slider);
                      
    button = gtk_toggle_button_new_with_label (TEXT_BUTTON_MAXIMUM_SPEED);
    g_signal_connect (G_OBJECT (button), "clicked", 
                      G_CALLBACK (button_pressed), "maximum_speed");
    gtk_box_pack_start (GTK_BOX (horizontalBox), button, FALSE, FALSE, 0);
    gtk_widget_show (button);

    button = gtk_toggle_button_new_with_label (TEXT_BUTTON_HISTOGRAM);
    g_signal_connect (G_OBJECT (button), "clicked", 
                      G_CALLBACK (button_pressed), "histogram");
    gtk_box_pack_start (GTK_BOX (horizontalBox), button, FALSE, FALSE, 0);
    gtk_widget_show (button);

    button = gtk_button_new_with_label (TEXT_BUTTON_PRESET);
    g_signal_connect (G_OBJECT (button), "clicked", 
                      G_CALLBACK (button_pressed), "preset");
    gtk_box_pack_start (GTK_BOX (horizontalBox), button, FALSE, FALSE, 0);
    gtk_widget_show (button);

#ifdef DISPLAY_RADIUS_SLIDER
    label = gtk_label_new (TEXT_LABEL_SLIDER_RADIUS);
    gtk_box_pack_start (GTK_BOX (horizontalBox), label, FALSE, FALSE, 0);
    gtk_widget_show (label);

    slider = gtk_hscale_new_with_range (MIN_BILLIARDRADIUS, 
                  MAX_BILLIARDRADIUS, 0.1);
    gtk_range_set_value (GTK_RANGE(slider), BILLIARDRADIUS);
    g_signal_connect (G_OBJECT (slider), "value_changed",
                      G_CALLBACK (slider_radius_changed), NULL);
    gtk_scale_set_draw_value (GTK_SCALE (slider), FALSE);
    gtk_widget_set_size_request (GTK_WIDGET (slider), 100, -1);
    gtk_range_set_update_policy (GTK_RANGE(slider), GTK_UPDATE_CONTINUOUS);
    gtk_box_pack_start (GTK_BOX (horizontalBox), slider, FALSE, FALSE, 0);
    gtk_widget_show (slider);
#endif

    button = gtk_button_new_with_label (TEXT_BUTTON_QUIT);
    g_signal_connect_swapped (G_OBJECT (button), "clicked", 
                      G_CALLBACK (gtk_main_quit), NULL);
    gtk_box_pack_start (GTK_BOX (horizontalBox), button, FALSE, FALSE, 0);
    gtk_widget_show (button);

    label = gtk_label_new ("");
    gtk_box_pack_start (GTK_BOX (horizontalBox), label, FALSE, FALSE, 0);
    gtk_widget_show (label);

    gtk_box_pack_start (GTK_BOX (verticalBox), horizontalBox, FALSE, FALSE, 10);
    gtk_widget_show (horizontalBox);

    gtk_container_add (GTK_CONTAINER (mainWindow), verticalBox);
    gtk_widget_show (verticalBox);

    gtk_window_set_resizable (GTK_WINDOW(mainWindow), FALSE);

    gtk_widget_show (mainWindow);

    histogram_set_up_colors ();
    billiard_set_up_screen (&leftBilliard);
    billiard_set_up_screen (&rightBilliard);
    initiate_redraw ();

    gtk_main ();

    return 0;
}
